GLOBAL.require "debugtools"
GLOBAL.require "actions"
GLOBAL.require "constants"
GLOBAL.require "recipe"
GLOBAL.require "components/playercontroller"

local is_DST = GLOBAL.MOD_API_VERSION == 10
local is_RoG = GLOBAL.SEASONS.AUTUMN ~= nil
local is_SW = GLOBAL.ACTIONS.HACK ~= nil
local is_vanilla = not is_DST and not is_RoG and not is_SW

if not is_DST or GLOBAL.TheNet:GetIsServer() then


HMM = {
}

local next = GLOBAL.next
local ACTIONS = GLOBAL.ACTIONS

local function AddActionButtonExtension(inst)
  local controller = inst 
  -- 1. Load configuration
  HMM.item_for = {}
  HMM.item_for[ACTIONS.CHOP] = "axe"
  HMM.item_for[ACTIONS.MINE] = "pickaxe"
  HMM.item_for[ACTIONS.DIG] = "shovel"
  HMM.item_for[ACTIONS.HAMMER] = "hammer"
  if is_SW then
    HMM.item_for[ACTIONS.HACK] = "machete"
  end
  HMM.hotkeys = {}
  
  HMM.hotkeys[ACTIONS.CHOP] = GLOBAL.KEY_U
  HMM.hotkeys[ACTIONS.MINE] = GLOBAL.KEY_I
  HMM.hotkeys[ACTIONS.DIG] = GLOBAL.KEY_O
  HMM.hotkeys[ACTIONS.HAMMER] = GLOBAL.KEY_P
  if is_SW then
    HMM.hotkeys[ACTIONS.HACK] = GLOBAL.KEY_Y
  end

  HMM.light = GetModConfigData("light")
  HMM.light_items = {"lighter","torch", "minerhat", 
    "lantern", "pumpkin_lantern", "yellowamulet",}
  HMM.player_emitters = {"lighterfire","torchfire", "minerhat", 
    "lantern", "pumpkin_lantern", "yellowamulet",}

  HMM.toggling = GetModConfigData("toggling")
  HMM.economic = GetModConfigData("economic")
  HMM.crafting = GetModConfigData("crafting")
  HMM.weapon = GetModConfigData("weapon")
  HMM.armor = GetModConfigData("armor")
  HMM.helmet = GetModConfigData("helmet")
  HMM.behaviour = {}
  HMM.behaviour[ACTIONS.CHOP] = GetModConfigData("chopping")
  HMM.behaviour[ACTIONS.MINE] = GetModConfigData("mining")
  HMM.behaviour[ACTIONS.DIG] = GetModConfigData("digging")
  HMM.behaviour[ACTIONS.HAMMER] = GetModConfigData("hammering")
  if is_SW then
    HMM.behaviour[ACTIONS.HACK] = GetModConfigData("hacking") 
  end 
  HMM.restore_item = nil
  HMM.has_restore_item = false

  HMM.actions_to_handle = {}
  for act, b in pairs(HMM.behaviour) do
    if b == 0 then
      HMM.item_for[act] = nil
    else
      table.insert(HMM.actions_to_handle, act)
    end
  end

  -- 2. Add hotkeys

  HMM.key_handler = {}
  if HMM.item_for[ACTIONS.CHOP] then
    HMM.key_handler[ACTIONS.CHOP] = function()
      HMM.actual_key_handler(ACTIONS.CHOP)
    end
  end
  if HMM.item_for[ACTIONS.DIG] then
    HMM.key_handler[ACTIONS.DIG] = function()
      HMM.actual_key_handler(ACTIONS.DIG)
    end
  end
  if HMM.item_for[ACTIONS.MINE] then
    HMM.key_handler[ACTIONS.MINE] = function()
      HMM.actual_key_handler(ACTIONS.MINE)
    end
  end
  if HMM.item_for[ACTIONS.HAMMER] then
    HMM.key_handler[ACTIONS.HAMMER] = function()
      HMM.actual_key_handler(ACTIONS.HAMMER)
    end
  end
  if is_SW and HMM.item_for[ACTIONS.HACK] then
    HMM.key_handler[ACTIONS.HACK] = function()
      HMM.actual_key_handler(ACTIONS.HACK)
    end
  end

  HMM.actual_key_handler = function(act)
    local player = is_DST and GLOBAL.ThePlayer or GLOBAL.GetPlayer()
    if not player then 
      return 
    end

    if player.components.playercontroller:IsEnabled() and 
      not GLOBAL.IsPaused() then
      HMM.single_action = act
      controller:DoActionButton()
      HMM.single_action = nil
    end
  end

  for act, key in pairs(HMM.hotkeys) do
    GLOBAL.TheInput:AddKeyDownHandler(key, HMM.key_handler[act])
  end


  -- 1. Darkness has to be handled from both Action and Attack buttons
  -- caller specified by parameter
  -- Return in_darkness, slot_with_light
  -- in_darkness: true if darkness not remedied (must try and craft)
  -- slot_with_light: slot used for lighting (must not be equipped with anything else)
  --TODO: Craft/equip while carrying torch in light of fire?
  controller.SortOutDarkness = function(self, is_attack)

    -- 1.1 Check if day/dusk/night and whether in darkness
    local in_darkness = self.inst.LightWatcher:GetLightValue() <=
      GLOBAL.TUNING.DARK_SPAWNCUTOFF
    local is_night = is_DST and GLOBAL.TheWorld.state.isnight
      or not is_DST and not GLOBAL.GetClock():IsDay()
    local is_cave = is_DST and GLOBAL.TheWorld:HasTag("cave")
      or not is_DST and GLOBAL.GetWorld():HasTag("cave")
    if not in_darkness and not (is_night or is_cave) then
      return false, nil
    end

    -- 1.2 Get equipped and active light sources
    local components = self.inst.components
    equipped = {}
    for _,v in pairs(GLOBAL.EQUIPSLOTS) do
      local i = components.inventory:GetEquippedItem(v)
      if i and 
	(not i.components.fueled or 
	i.components.fueled and i.components.fueled.currentfuel > 0) and
	table.contains(HMM.light_items, i.prefab) then 
	equipped[v] = i
      end
    end

    -- 1.2 Deal with darkness
    -- Torch (e.g.) takes a moment between being equipped and lighting up,
    -- hence check whether a light source is equipped or not
    if in_darkness and next(equipped) == nil then
      -- 1.2.2 Get all equippable items, return if none
      local in_inv = components.inventory:FindItems(function(i) 
	return (not i.components.fueled or 
	i.components.fueled and i.components.fueled.currentfuel > 0) and
	table.contains(HMM.light_items, i.prefab)
      end)
      equippable = {}
      for _,v in pairs(in_inv) do
	equippable[v.prefab] = v
      end
      if next(equippable) == nil then
	return true, nil
      end
      -- 1.2.3 Equip according to priority rules
      if is_attack then
	local x, y, z = self.inst.Transform:GetWorldPosition()
	local pos = GLOBAL.Vector3(x + 0.5, y, z + 0.5)
	if equippable.pumpkin_lantern then 
	  components.inventory:DropItem(equippable.pumpkin_lantern,
	    false, true, pos) 
	  return false, nil
	elseif equippable.lantern then
	  components.inventory:DropItem(equippable.lantern,
	    false, true, pos) 
	  return false, nil
	elseif equippable.minerhat then
	  components.inventory:Equip(equippable.minerhat) 
	  return false, GLOBAL.EQUIPSLOTS.HEAD
	elseif equippable.yellowamulet then
	  components.inventory:Equip(equippable.yellowamulet) 
	  return false, GLOBAL.EQUIPSLOTS.BODY
	elseif equippable.torch then
	  components.inventory:Equip(equippable.torch) 
	  return false, GLOBAL.EQUIPSLOTS.HANDS
	elseif equippable.lighter then
	  components.inventory:Equip(equippable.lighter) 
	  return false, GLOBAL.EQUIPSLOTS.HANDS
	else
	  return true, nil
	end
      else
	if equippable.minerhat then
	  components.inventory:Equip(equippable.minerhat) 
	  return false, nil
	elseif equippable.lantern then
	  components.inventory:Equip(equippable.lantern)
	  return false, GLOBAL.EQUIPSLOTS.HANDS
	elseif equippable.torch then
	  components.inventory:Equip(equippable.torch) 
	  return false, GLOBAL.EQUIPSLOTS.HANDS
	elseif equippable.lighter then
	  components.inventory:Equip(equippable.lighter) 
	  return false, GLOBAL.EQUIPSLOTS.HANDS
	else
	  return true, nil
	end
      end
    else -- not in_darkness or equipped 
      -- 1.2.4 Find whether light source can be safely put away
      if next(equipped) ~= nil then
	local x, y, z = self.inst.Transform:GetWorldPosition()
	local nearby = GLOBAL.TheSim and
	  GLOBAL.TheSim:FindEntities(x, y, z, 20, {"lightsource"}, {"INLIMBO"}) 
	if next(nearby) ~= nil then 
	  local n = 0
	  for k,v in pairs(nearby) do
	    local distsq = self.inst:GetDistanceSqToInst(v)
	    local sqradius = 0.81 * v.Light:GetCalculatedRadius()
	    if sqradius > distsq and
	      (not table.contains(HMM.player_emitters, v.prefab) or  
		(v.prefab == "lantern" and 
		not v.components.equippable.isequipped)) then
		return false, nil
	    end
	  end
	end
	-- 1.2.5 If safe to put away, choose which one from possibly multiple
	local unequip_priority = {
	  GLOBAL.EQUIPSLOTS.HEAD,
	  GLOBAL.EQUIPSLOTS.BODY,
	  GLOBAL.EQUIPSLOTS.HANDS}
	for _,v in pairs(unequip_priority) do
	  for _,u in pairs(equipped) do
	    if u.components.equippable.equipslot == v then
	      return false, v
	    end
	  end
	end
      end
      return false, nil --unreachable
    end
  end


  -- 2. Add functionality to attack button

  controller.DoAttackButton_before_handymanness =
    controller.DoAttackButton
  controller.DoAttackButton = function(self)
    local res = self:DoAttackButton_before_handymanness()
    local in_darkness, slot_with_light = self:SortOutDarkness(true) 
    local components = is_DST and self.inst.components 
      or self.inst.components
    if components.combat.target then
      local hm = components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.HEAD)
      -- 2.1 Find best weapon to equip
      if HMM.weapon and
	(not slot_with_light or slot_with_light ~= GLOBAL.EQUIPSLOTS.HANDS) then
	local wp = components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.HANDS)
	local curr_damage = wp and wp.components.weapon and
	  wp.components.weapon.damage or 0
	local min_damage = curr_damage
	local to_equip = components.inventory:FindItems(function(i) 
	  local ret = i.components.weapon and 
	    (HMM.weapon == 2 or not i.components.lighter) and
	    min_damage < i.components.weapon.damage
	  if ret then min_damage = i.components.weapon.damage end
	  return ret
	  end)
	for _, v in pairs(to_equip) do
	  if v.components.weapon.damage == min_damage and
	    v.components.weapon.damage > curr_damage then
	    components.inventory:Equip(v) 
	    break
	  end
	end
      end
      -- 2.2 Find any armor/helmet
      local DoArmor = function(config, slot_with_light, target_slot)
	if config and slot_with_light ~= target_slot then
	  local ar = components.inventory:GetEquippedItem(target_slot)
	  if ar and ar.components.armor then return end
	  local to_equip = components.inventory:FindItem(function(i) 
	    return i.components.armor and 
	      i.components.equippable.equipslot == target_slot
	  end)
	  if to_equip then components.inventory:Equip(to_equip) end
	end
      end
      DoArmor(HMM.armor, slot_with_light, GLOBAL.EQUIPSLOTS.BODY)
      DoArmor(HMM.helmet, slot_with_light, GLOBAL.EQUIPSLOTS.HEAD)
    end
  end
  

  controller.OnUpdate_before_handymanness = controller.OnUpdate

  controller.OnUpdate = function(self, dt)
    self:OnUpdate_before_handymanness(dt)
    -- do automagic control repeats
    if self.inst.sg:HasStateTag("idle") then
      for act, key in pairs(HMM.hotkeys) do
	if GLOBAL.TheInput:IsControlPressed(key) then
	  HMM.actual_key_handler(key, act)
	  break
	end
      end
    end
  end

  -- 4. Helper functions
  controller.RestoreSavedItem = function(self)
    if HMM.has_restore_item then
      HMM.has_restore_item = false
      local components = is_DST and self.inst.components 
	or self.inst.components
      components.inventory:Equip(HMM.restore_item)
    end
  end

  controller.IsValidTarget = function(self, act, tgt, force)
    local workable = tgt.components.workable
    local hackable = nil
    if is_SW then
      hackable = tgt.components.hackable
    end
    if workable and workable.action == act or
      is_SW and hackable and 
      act == ACTIONS.HACK and hackable.canbehacked then

      local good = force or self:IsGoodTarget(act, tgt)
      if is_SW and good and act == ACTIONS.DIG and
	tgt.prefab == "treasurechest" then
	good = false
      end
      return good
    end
  end

  controller.IsGoodTarget = function(self, act, tgt)
    local ret = true
    if act == ACTIONS.HAMMER then
      if not string.find(tgt.prefab, "flotsam_") and
	not string.find(tgt.prefab, "wallyintro_") and
	not string.find(tgt.prefab, "skeleton") and
	not string.find(tgt.prefab, "wildboarhead") and
	not string.find(tgt.prefab, "crate") then
	ret = false
      end
    elseif act == ACTIONS.MINE then
      if tgt.prefab == "limpetrock" then
	ret = false
      end
    elseif act == ACTIONS.DIG then
      if string.find(tgt.prefab, "hole") or
	string.find(tgt.prefab, "sandhill") or
	tgt.components.pickable or
	tgt.components.hackable then
	ret = false
      end
    elseif act == ACTIONS.CHOP then
      if string.find(tgt.prefab, "evergreen") then
	ret = not tgt.components.growable or
	  tgt.components.growable.stage == 3 
      else
	ret = not tgt.components.growable or
	  tgt.components.growable.stage == 
	  #tgt.components.growable.stages 
      end
    end
    return ret
  end

  controller.FindPossibilities = function(self, action_list, economic)
    local components = self.inst.components
    local inventory = components.inventory:FindItems(function(i) 
      return i.components.tool ~= nil end)

    local force_action = {}
    local equipped = 
      components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.HANDS)
    if equipped and equipped.components.tool then
      for _, act in pairs(action_list) do
	if equipped.components.tool:CanDoAction(act) then
	  table.insert(force_action,act)
	end
      end
    end

    local possibilities = {}
    for _, act in pairs(action_list) do
      for _, tool in pairs(inventory) do
	if tool.components.tool:CanDoAction(act) then
	  possibilities[act] = {equippable = tool}
	  break
	end
      end
      if not possibilities[act] then
	local item = HMM.item_for[act]
	if not components.builder:CanBuild(item) or
	  not components.builder:KnowsRecipe(item) then item = nil end
	local golden = "golden" .. HMM.item_for[act]
	if not components.builder:CanBuild(golden) or
	  not components.builder:KnowsRecipe(golden) then golden = nil end
	if HMM.crafting == 3 and golden or
	  HMM.crafting == 2 and not item then
	  
	  item = golden
	end
	if item then
	  if economic then
	    for _, o in pairs(economic) do
	      if o.components.tool and o.components.tool:CanDoAction(act) then
		possibilities[act] = {economic=o}
		break
	      end
	    end
	  end
	  if not possibilities[act] then
	    possibilities[act] = {craftable=item}
	  end
	end
      end
    end
    for _, act in pairs(force_action) do
      possibilities[act] = nil
    end
    return possibilities, force_action
  end

  -- 5. Add functionality to action button

  controller.DoActionButton_before_handymanness =
    controller.DoActionButton

  -- original returns an unused boolean, happens to be always true...
  -- But since it is never checked, I've used it for some debugging purposes
  -- Two wrongs make a bigger wrong.
  controller.DoActionButton = function(self)
    -- 5.1 Setup modifiers etc
    local components = self.inst.components
    local alt = GLOBAL.TheInput:IsKeyDown(GLOBAL.KEY_ALT) or
      GLOBAL.TheInput:IsKeyDown(GLOBAL.KEY_PERIOD)
    local ctrl = GLOBAL.TheInput:IsKeyDown(GLOBAL.KEY_CTRL)
    local inhibit = (alt and HMM.toggling == 0) or 
      (ctrl and HMM.toggling == 1)
    local force = (ctrl and HMM.toggling == 0) or 
      (alt and HMM.toggling == 1)
    --
    -- 5.2 Handle darkness first
    local in_darkness, slot_with_light = false, nil
    if not inhibit then 
      in_darkness, slot_with_light = self:SortOutDarkness(false) 
      if in_darkness then
	for k,v in pairs(HMM.light_items) do
	  if components.builder:CanBuild(v) then
	    local light_recipe = is_DST and GLOBAL.GetValidRecipe(v)
	      or GLOBAL.GetRecipe(v)
	    return light_recipe and 
	      (is_DST and 
	        components.builder:MakeRecipeFromMenu(light_recipe) or
	        components.builder:MakeRecipe(light_recipe))
	  end
	end
      end
    end
    
    -- 5.3 "Old" behaviour
    local x, y, z = self.inst.Transform:GetWorldPosition()
    local to_block = {}
    local nearby = GLOBAL.TheSim and GLOBAL.TheSim:FindEntities(x, y, z, 6, {}, {"INLIMBO"})

    local equipped = components.inventory:GetEquippedItem(GLOBAL.EQUIPSLOTS.HANDS)
    if equipped and not equipped.components.tool then
	equipped = nil
    end
    
    if HMM.single_action then
      if equipped and 
	equipped.components.tool.action == single_action then
	for _, o in pairs(nearby) do
	  if o.components.workable and 
	    (o.components.workable.action == HMM.single_action and
	    force or self:IsGoodTarget(act, o)) then

	    components.locomotor:PushAction(GLOBAL.BufferedAction(self.inst,
	      o, single_action), true) 
	    return true
	  end
	end
	return false
      end
    else
      if (not is_vanilla) then
	if not force then
	  for _, o in pairs(nearby) do
	    if o.components.workable then
	      local act = o.components.workable.action
	      if not force and not self:IsGoodTarget(act, o) then

		table.insert(to_block, o)
	      end
	    end
	  end
	  for _, o in pairs(to_block) do
	    if o.components.workable then
	      o.components.workable:SetWorkable(false)
	    end
	  end
	end
      end
	
      --TODO: must check or can use return value?
      local res = self:DoActionButton_before_handymanness()

      if (not is_vanilla) then
	if not force then
	  for _, o in pairs(to_block) do
	    if o.components.workable then
	      o.components.workable:SetWorkable(true)
	    end
	  end
	end
      end

      if components.locomotor.bufferedaction or
	components.locomotor:HasDestination() then
	return true
      end
    end

    -- 5.4 Stop here if inhibited or hands busy (with light)
    -- or active item
    local active_item = components.inventory:GetActiveItem() 
    if inhibit or active_item or
      slot_with_light == GLOBAL.EQUIPSLOTS.HANDS then
      return false
    end

    -- 5.5 What can be done?
    local actions_to_handle = 
      HMM.single_action and {HMM.single_action} or
      HMM.actions_to_handle

    local economic = HMM.economic and 
      GLOBAL.TheSim:FindEntities(x, y, z, 20, {}, {"INLIMBO"})
    local possibilities, force_action = 
      self:FindPossibilities(actions_to_handle, economic)


    -- 5.6 What targets to look for, what tools to equip/craft
    
    if next(force_action) ~= nil then
      for _, act in pairs(force_action) do
	for _, tgt in pairs(nearby) do
	  if tgt.components.workable and self:IsValidTarget(act, tgt, force) then
	    components.locomotor:PushAction(GLOBAL.BufferedAction(self.inst,
	      tgt, act), true) 
	    return true
	  end
	end
      end
    end

    if not possibilities then
      self:RestoreSavedItem()
      return true
    end

    for _, tgt in pairs(nearby) do
      for act, what in pairs(possibilities) do
	if tgt.components.workable and self:IsValidTarget(act, tgt, force) then
	  if what.equippable then
	    if not HMM.has_restore_item then
	      HMM.restore_item = equipped
	      HMM.has_restore_item = HMM.restore_item ~=nil
	    end
	    if components.inventory:Equip(what.equippable) then
	      return true
	    end
	  elseif what.craftable then
	    local recipe = is_DST and 
	      GLOBAL.GetValidRecipe(what.craftable) or
	      GLOBAL.GetRecipe(what.craftable)

	      return is_DST and 
		components.builder:MakeRecipeFromMenu(recipe) or
		components.builder:MakeRecipe(recipe)
	  elseif what.economic then
	      print("Say something about ", what.economic, "?")
	  end
	end
      end
    end

  end

  return true
end

AddClassPostConstruct("components/playercontroller", AddActionButtonExtension)

end --if not is_DST or is server


